/*---------------------------------------------------------------------------------------------
 *  Copyright (c) Microsoft Corporation. All rights reserved.
 *  Licensed under the MIT License. See License.txt in the project root for license information.
 *--------------------------------------------------------------------------------------------*/
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { onUnexpectedExternalError } from '../../../base/common/errors.js';
import { Emitter } from '../../../base/common/event.js';
import { DisposableStore } from '../../../base/common/lifecycle.js';
import { Position } from '../../common/core/position.js';
import { Range } from '../../common/core/range.js';
import { InlayHintsProviderRegistry } from '../../common/languages.js';
export class InlayHintAnchor {
    constructor(range, direction, usesWordRange) {
        this.range = range;
        this.direction = direction;
        this.usesWordRange = usesWordRange;
    }
}
export class InlayHintItem {
    constructor(hint, anchor, _provider) {
        this.hint = hint;
        this.anchor = anchor;
        this._provider = _provider;
        this._isResolved = false;
    }
    with(delta) {
        const result = new InlayHintItem(this.hint, delta.anchor, this._provider);
        result._isResolved = this._isResolved;
        result._currentResolve = this._currentResolve;
        return result;
    }
    resolve(token) {
        return __awaiter(this, void 0, void 0, function* () {
            if (typeof this._provider.resolveInlayHint !== 'function') {
                return;
            }
            if (this._currentResolve) {
                // wait for an active resolve operation and try again
                // when that's done.
                yield this._currentResolve;
                if (token.isCancellationRequested) {
                    return;
                }
                return this.resolve(token);
            }
            if (!this._isResolved) {
                this._currentResolve = this._doResolve(token)
                    .finally(() => this._currentResolve = undefined);
            }
            yield this._currentResolve;
        });
    }
    _doResolve(token) {
        var _a, _b;
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const newHint = yield Promise.resolve(this._provider.resolveInlayHint(this.hint, token));
                this.hint.tooltip = (_a = newHint === null || newHint === void 0 ? void 0 : newHint.tooltip) !== null && _a !== void 0 ? _a : this.hint.tooltip;
                this.hint.label = (_b = newHint === null || newHint === void 0 ? void 0 : newHint.label) !== null && _b !== void 0 ? _b : this.hint.label;
                this._isResolved = true;
            }
            catch (err) {
                onUnexpectedExternalError(err);
                this._isResolved = false;
            }
        });
    }
}
export class InlayHintsFragments {
    constructor(data, model) {
        this._disposables = new DisposableStore();
        this._onDidChange = new Emitter();
        this.onDidReceiveProviderSignal = this._onDidChange.event;
        const items = [];
        for (const [list, provider] of data) {
            this._disposables.add(list);
            for (let hint of list.hints) {
                // compute the range to which the item should be attached to
                let position = hint.position;
                let direction = 'before';
                let range = Range.fromPositions(position);
                let word = model.getWordAtPosition(position);
                let usesWordRange = false;
                if (word) {
                    if (word.endColumn === position.column) {
                        direction = 'after';
                        usesWordRange = true;
                        range = wordToRange(word, position.lineNumber);
                    }
                    else if (word.startColumn === position.column) {
                        usesWordRange = true;
                        range = wordToRange(word, position.lineNumber);
                    }
                }
                items.push(new InlayHintItem(hint, new InlayHintAnchor(range, direction, usesWordRange), provider));
            }
            if (provider.onDidChangeInlayHints) {
                provider.onDidChangeInlayHints(this._onDidChange.fire, this._onDidChange, this._disposables);
            }
        }
        this.items = items.sort((a, b) => Position.compare(a.hint.position, b.hint.position));
    }
    static create(model, ranges, token) {
        return __awaiter(this, void 0, void 0, function* () {
            const data = [];
            const promises = InlayHintsProviderRegistry.ordered(model).reverse().map(provider => ranges.map((range) => __awaiter(this, void 0, void 0, function* () {
                try {
                    const result = yield provider.provideInlayHints(model, range, token);
                    if (result === null || result === void 0 ? void 0 : result.hints.length) {
                        data.push([result, provider]);
                    }
                }
                catch (err) {
                    onUnexpectedExternalError(err);
                }
            })));
            yield Promise.all(promises.flat());
            return new InlayHintsFragments(data, model);
        });
    }
    dispose() {
        this._onDidChange.dispose();
        this._disposables.dispose();
    }
}
function wordToRange(word, lineNumber) {
    return new Range(lineNumber, word.startColumn, lineNumber, word.endColumn);
}
